#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-23 20:45:03
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-23 20:49:59
"""

from enum import Enum


class TimeUnit(Enum):
    fs = "fs"
    ps = "ps"
    ns = "ns"
    us = "us"
    ms = "ms"
    s = "s"

    @staticmethod
    def fromString(timeUnitStr: str):
        if timeUnitStr == "fs":
            return TimeUnit.fs
        elif timeUnitStr == "ps":
            return TimeUnit.ps
        elif timeUnitStr == "ns":
            return TimeUnit.ns
        elif timeUnitStr == "us":
            return TimeUnit.us
        elif timeUnitStr == "ms":
            return TimeUnit.ms
        elif timeUnitStr == "s":
            return TimeUnit.s
        else:
            return None

    @staticmethod
    def toString(timeUnit: "TimeUnit"):
        return timeUnit.value if timeUnit is not None else None


class TimeValue:
    def __init__(self, value: float, unit: TimeUnit) -> None:
        self.value = value
        self.unit = unit

    def __repr__(self) -> str:
        return f"{int(self.value)} {self.unit}"



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-23 23:59:53
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-24 01:11:59
"""

from typing import Any


class Range:
    def __init__(self, start: int, end: int = None):
        self.start = start
        self.end = end


def rangeToString(range: Range) -> str:
    if range is None:
        return ""
    assert isinstance(range, Range), f"Expected Range, got {type(range)}"
    return (
        f"[{range.start}:{range.end}]" if range.end is not None else f"[{range.start}]"
    )


class DFGNode:
    # Expressions are hard to handle because of the recursive nature

    def __init__(self, name: str = None) -> None:
        self.variable_name = name
        self.operation = None
        self.range = None
        self.children = []

    def __repr__(self) -> str:
        return f"Node({self.variable_name})"

    def isVariable(self):
        return self.operation == OPType.VARIABLE

    def isConstant(self):
        return self.operation == OPType.CONSTANT

    def setRange(self, range: Range):
        self.range = range

    def toString(self) -> str:
        return (
            self.variable_name
            if self.range is None
            else f"{self.variable_name}{rangeToString(self.range)}"
        )

    @property
    def name(self):
        return self.variable_name

    @property
    def needsParentheses(self):
        return True

    def replaceVariable(self, old: str, new: str):
        if self.variable_name == old:
            self.variable_name = new
        child: DFGNode
        for child in self.children:
            child.replaceVariable(old, new)

    def copy(self):
        node = DFGNode(self.variable_name)
        node.operation = self.operation
        node.range = self.range
        for child in self.children:
            node.children.append(child.copy())
        return node


from enum import Enum


class OPType(Enum):
    BINARY_AND = "binary_and"
    BINARY_OR = "binary_or"
    BINARY_BITAND = "binary_bitand"
    BINARY_XOR = "binary_xor"
    BINARY_XNOR = "binary_xnor"
    BINARY_BITOR = "binary_bitor"
    BINARY_EQ = "binary_eq"
    BINARY_NEQ = "binary_neq"
    BINARY_EQ_EXT = "binary_eq_ext"
    BINARY_NEQ_EXT = "binary_neq_ext"
    BINARY_LT = "binary_lt"
    BINARY_GT = "binary_gt"
    BINARY_LEQ = "binary_leq"
    BINARY_GEQ = "binary_geq"
    BINARY_RSHIFT = "binary_rshift"
    BINARY_RSHIFT_EXT = "binary_rshift_ext"
    BINARY_LSHIFT = "binary_lshift"
    BINARY_LSHIFT_EXT = "binary_lshift_ext"

    # Arithmetic operations
    BINARY_ADD = "binary_add"
    BINARY_SUB = "binary_sub"
    BINARY_MUL = "binary_mul"
    BINARY_DIV = "binary_div"
    BINARY_MOD = "binary_mod"
    BINARY_POW = "binary_pow"

    # Unary operations
    UNARY_POS = "unary_pos"
    UNARY_NEG = "unary_neg"
    UNARY_NOT = "unary_not"
    UNARY_INV = "unary_inv"
    UNARY_AND = "unary_and"
    UNARY_OR = "unary_or"
    UNARY_XOR = "unary_xor"
    UNARY_NAND = "unary_nand"
    UNARY_NOR = "unary_nor"
    UNARY_XNOR = "unary_xnor"
    CONDITIONAL_EXPRESSION = "conditional_expression"

    VARIABLE = "variable"
    CONSTANT = "constant"
    MACRO = "macro"

    ARRAY_CONCAT = "array_concat"
    ARRAY_REPLICATE = "array_replicate"
    ARRAY_SLICE = "array_slice"
    ARRAY_INDEX = "array_index"

    FUNCTION_CALL = "function_call"

    EVENT_INIT = "event_init"
    EVENT_COMB = "event_comb"
    EVENT_POSEDGE = "event_posedge"
    EVENT_NEGEDGE = "event_negedge"

    EVENT_AND = "event_and"
    EVENT_OR = "event_or"

    UNKNOWN = "unknown"


class OPNode(DFGNode):
    def __init__(self, op: str, operation: OPType, *children) -> None:
        super().__init__(op)
        self.operation = operation
        self.children = children


class VarNode(DFGNode):
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.operation = OPType.VARIABLE


class ConstantNode(DFGNode):
    def __init__(self, value: Any) -> None:
        super().__init__(str(value))
        self.operation = OPType.CONSTANT


class EmptyEvent(DFGNode):
    def __init__(self) -> None:
        super().__init__("*")
        self.operation = OPType.EVENT_COMB


class InitEvent(DFGNode):
    def __init__(self) -> None:
        super().__init__("*")
        self.operation = OPType.EVENT_INIT


class UndefinedNode(DFGNode):
    def __init__(self) -> None:
        super().__init__("*")
        self.operation = OPType.UNKNOWN



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 23:50:49
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:58:28
"""

from lark import Transformer, v_args, Tree
from ...modules import *
from .larkUtils import *


class PortTransformer(Transformer):

    # port_range : "[" expression ":" expression "]"
    @v_args(inline=True)
    def port_range(self, start, end):
        assert isinstance(start, DFGNode), f"start = {start}, type = {type(start)}"
        assert isinstance(end, DFGNode), f"end = {end}, type = {type(end)}"
        return Range(start, end)

    def variable_list(self, items):
        return items

    # declaration: [port_direction] [port_type] [port_range] variable_list ["=" expression] ";"
    def declaration(self, items):
        statements = []
        dir, t, range, variables, expression = items
        dir = PortDirection.fromString(dir.data) if dir is not None else None
        t = PortType.fromString(t.data) if t is not None else None
        ports = []
        for variable in variables:
            port = Port(variable)
            port.setAll(dir, t, range)
            if expression:
                assert len(variables) == 1
                statements.append(BlockingAssignment(variable, expression))
            ports.append(port)
        statements.append(PortDeclaration(ports))
        return statements



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 18:22:19
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 19:03:09
"""

import logging

logger = logging.getLogger(__name__)

from .dfg import *


class Parameter:
    def __init__(self, range: Range, name: str, value: str):
        logger.info(f"Parameter: {name} = {value}")
        self.name = name
        self.range = range
        self.value = value

    def __repr__(self):
        return "#{}".format(self.name, self.value)

    def getName(self):
        return self.name


class ParameterHandler:
    def __init__(self):
        self.param_list = {}

    def addParameter(self, param: Parameter):
        self.param_list[param.getName()] = param

    def addParameters(self, parameters: list):
        if not parameters:
            return
        for parameter in parameters:
            self.addParameter(parameter)

    def getParameter(self, name: str):
        return self.param_list[name]

    def getParameters(self):
        return self.param_list



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-25 14:55:16
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-25 15:59:34
"""


class ModuleInst:
    def __init__(self, name: str) -> None:
        self.name = name
        self.module_name = None
        self.port_list = []
        self.port_names = {}
        self.inplace_parameters = None
        self.parameter_list = {}

        # self.port_index: int = 0
        self.is_standard_inst = True

    def addPort(self, name: str, value: str):
        self.port_list.append(value)
        index = len(self.port_list) - 1
        if name is None:
            name = f"port_{index}"
            self.is_standard_inst = False
        self.port_names[name] = index

    def getPort(self, name: str):
        index = self.port_names[name]
        return self.port_list[index]

    def getPorts(self):
        return self.port_list

    def getPortNames(self):
        return self.port_list

    def addParameter(self, name: str, value: str):
        self.parameter_list[name] = value

    def setModuleName(self, name: str):
        self.module_name = name

    def setInplaceParameters(self, inplace_parameters):
        self.inplace_parameters = inplace_parameters[:]

    def getInplaceParameters(self):
        return self.inplace_parameters

    def hasInplaceParameters(self):
        return self.inplace_parameters is not None

    def getInstName(self):
        return self.name

    def getModuleName(self):
        return self.module_name

    def getPortList(self):
        return self.port_names.keys()

    def getParameters(self):
        return self.parameter_list

    def hasParameters(self):
        return len(self.parameter_list) > 0



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-28 23:10:58
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-30 12:44:07
"""

from .verilog import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-23 22:34:21
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-23 22:34:55
"""

import datetime


def writeHeader(f):
    headerString = ""
    headerString += "//" + "-" * 78 + "\n"
    headerString += "// " + "Generated by LeaP compiler" + "\n"
    headerString += "// " + "Author: " + "Hanyu Wang" + "\n"
    headerString += (
        "// " + "Date: " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + "\n"
    )
    headerString += "//" + "-" * 78 + "\n"
    f.write(headerString)



from frontend.verilog.modules.dfg import Any
from frontend.verilog.modules.module import Any, Module
from .dfg import *
from .module import *
from .moduleInst import *


class Statement:
    """
    Statement is the base class for all statements in the module.
    It is a callable object that can be called with a module object to add the statement to the module.
    """

    def __init__(self) -> None:
        self.condition: DFGNode = None
        self.event: DFGNode = None

    def setCondition(self, condition):
        self.condition = condition

    def addCondition(self, condition: DFGNode):
        if condition is None:
            return
        if self.condition is None:
            self.condition = condition
        else:
            self.condition = OPNode("&&", OPType.BINARY_AND, self.condition, condition)
        return self

    def setEvent(self, event: DFGNode):
        self.event = event
        return self

    def __call__(self, module: Module) -> Any:
        raise NotImplementedError


class ModuleInstantiation(Statement):
    def __init__(self, inst: ModuleInst) -> None:
        super().__init__()
        self.inst = inst

    def __call__(self, module: Module) -> Any:
        module.addInstance(self.inst)

    def __repr__(self) -> str:
        return f"ModuleInstantiation({self.inst})"


class PortDeclaration(Statement):
    def __init__(self, ports: list) -> None:
        super().__init__()
        self.ports = ports

    def __repr__(self) -> str:
        return f"PortDeclaration({self.ports})"

    def __call__(self, module: Module) -> Any:
        module.addPorts(self.ports)


class MacroDefinition(Statement):
    def __init__(self, name: str, value: DFGNode) -> None:
        super().__init__()
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return f"MacroDefinition({self.name} = {self.value})"

    def __call__(self, module: Module) -> Any:
        module.addMacro(self.name, self.value)


class Assignment(Statement):
    def __init__(
        self, target: DFGNode, expression: DFGNode, condition: DFGNode = None
    ) -> None:
        self.target: DFGNode = target
        self.expression: DFGNode = expression
        self.condition: DFGNode = condition
        self.type = None
        self.isBlocking = False

    def __repr__(self) -> str:
        retString = ""
        if self.isBlocking:
            retString += f"{self.target} = {self.expression}"
        else:
            retString += f"{self.target} <= {self.expression} ({self.condition})"
        return retString

    def replaceVariable(self, old: str, new: str):
        self.target.replaceVariable(old, new)
        self.expression.replaceVariable(old, new)
        if self.condition is not None:
            self.condition.replaceVariable(old, new)

    def __call__(self, module: Module) -> Any:
        return module.addAssignment(self)


class BlockingAssignment(Assignment):
    def __init__(
        self, target: DFGNode, expression: DFGNode, condition: DFGNode = None
    ) -> None:
        super().__init__(target, expression, condition)
        self.isBlocking = True


class NonBlockingAssignment(Assignment):
    def __init__(
        self, target: DFGNode, expression: DFGNode, condition: DFGNode = None
    ) -> None:
        super().__init__(target, expression, condition)
        self.isBlocking = False



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-19 02:24:23
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-19 02:31:37
"""

import re


def flattenVerilog(inputFile: str, outputFile: str):
    with open(inputFile, "r") as f:
        lines = f.readlines()

    with open(outputFile, "w") as f:
        for line in lines:
            # we look for "[x]" and replace it with "_idx_x_"

            m = re.search(r"\[(\d+)\]", line)
            if m is not None:
                line = re.sub(r"\[(\d+)\]", r"_idx_\1_", line)

            # and we look for "." and replace it with "_dot_"
            line = re.sub(r"\.", r"_dot_", line)

            f.write(line)



from lark import Transformer, v_args
from ...modules import *
from .larkUtils import *


class EventTransformer(Transformer):

    def always_block(self, items):
        event_condition: DFGNode = items[0]
        assert isinstance(event_condition, DFGNode)
        return [x.setEvent(event_condition) for x in items[1]]

    def initial_block(self, items):
        event_condition: DFGNode = InitEvent()
        assert isinstance(event_condition, DFGNode)
        return [x.setEvent(event_condition) for x in items[0]]

    def combinational_event_expression(self, _):
        return EmptyEvent()

    def posedge_event_expression(self, items):
        return OPNode("posedge", OPType.EVENT_POSEDGE, items[0])

    def negedge_event_expression(self, items):
        return OPNode("negedge", OPType.EVENT_NEGEDGE, items[0])

    def event_or(self, items):
        return OPNode("or", OPType.EVENT_OR, items)

    def event_and(self, items):
        return OPNode("and", OPType.EVENT_AND, items)



from ..modules import *
import re


def extractDataFlowNodesRec(graph: pgv.AGraph, node: pgv.Node, visited: set):
    if node in visited:
        return
    visited.add(node)
    for edge in graph.in_edges(node):
        # do not traverse condition
        if graph.get_edge(edge[0], edge[1]).attr["style"] != "dashed":
            extractDataFlowNodesRec(graph, edge[0], visited)


def extractDataFlowControlFlowNodes(graph: pgv.AGraph, dataOutputs: list):

    dataNodes = set()
    controlNodes = []
    for out in dataOutputs:
        extractDataFlowNodesRec(graph, out, dataNodes)
    for node in graph.nodes():
        if node not in dataNodes:
            controlNodes.append(node)
    return dataNodes, controlNodes


def extractDataFlowControlFlow(CDFG: pgv.AGraph, dataOutputs: list):
    dataNodes, controlNodes = extractDataFlowControlFlowNodes(CDFG, dataOutputs)
    dataGraph = CDFG.subgraph(dataNodes)
    controlGraph = CDFG.subgraph(controlNodes)
    return dataGraph, controlGraph



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 23:17:28
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:51:25
"""

from .systemVerilogTransformer import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-05-17 15:08:32
Last Modified by: Hanyu Wang
Last Modified time: 2024-05-17 15:08:34
"""

from .modules import *
from .reader import *
from .writer import *
from .fixVerilog import *
from .writer import *
from .changeName import *
from .algorithms import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-23 22:32:42
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-23 23:20:36
"""

from .writer import *
from .dfgWriter import *
from .headerWriter import *



import lark


def test_00_lark():
    grammar = """
    start: expression

    ?expression: NAME
                | "(" expression ")"
                | expression "==" expression -> eq
                | "!" expression -> not
    
    %import common.CNAME -> NAME
    %import common.WS
    %ignore WS
    """
    parser = lark.Lark(grammar, parser="earley", lexer="dynamic_complete")
    tree = parser.parse("!(a == b)")
    tree = parser.parse("!a == b")
    print(tree.pretty())


if __name__ == "__main__":
    test_00_lark()



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 18:21:53
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:25:03
"""

from .parameters import *
from .ports import *
from .module import *
from .netlist import *
from .dfg import *
from .timeUnit import *
from .moduleInst import *
from .statement import *



## Assumptions:
# 1. The name of the "end" signal is known (default: "finish")
# 2. The name of the "reset" signal is known (default: "reset")

if __name__ == "__main__":
    from frontend import *

    parameters = {
        "extractFSM": False,
        "extractCIP": False,
        "extractARG": False,
        "extractREG": False,
    }

    network: Netlist = readVerilog("examples/verilogs/external/legup.v")
    module = network.getModule("toy")
    finishPort = "finish"
    outputsNames = [
        port
        for port in module.getPortsByType(PortDirection.OUTPUT)
        if port != finishPort
    ]
    CDFG = module.getGraph()
    DFG, CFG = extractDataFlowControlFlow(CDFG, outputsNames)
    module.separateCDFG(DFG, CFG)
    # module.exportDOT(params=parameters)
    # module.exportDOT()
    module.writeDOT("toy.dot")



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 19:20:27
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:51:31
"""

import logging

logger = logging.getLogger(__name__)

from ...modules import *
from .expressionTransformer import *
from .conditionTransformer import *
from .portTransformer import *
from .eventTransformer import *
from .moduleTransformer import *


class SystemVerilogTransformer(
    ModuleTransformer,
    ConditionTransformer,
    PortTransformer,
    ExpressionTransformer,
    EventTransformer,
):
    # description: (macro_definition | timescale_directive | module)+
    def start(self, items):
        modules = []
        definitions = {}
        for item in items:
            if isinstance(item, Module):
                modules.append(item)
            elif isinstance(item, dict):
                for key, value in item.items():
                    key = str(key)
                    value = str(value)
                    definitions[key] = value
        return Netlist(modules, definitions)

    # timescale_directive: "`timescale" time_unit "/" time_precision
    # time_precision: NUMBER time_identifier
    # time_identifier: "s" | "ms" | "us" | "ns" | "ps"
    def timescale_directive(self, items):
        time_unit = items[0]
        time_precision = items[1]
        return {"time_unit": time_unit, "time_precision": time_precision}

    # time_unit: NUMBER time_identifier
    def time_unit(self, items):
        return TimeValue(float(items[0]), items[1].data)

    def time_precision(self, items):
        return TimeValue(float(items[0]), items[1].data)

    # macro_definition: "`define" identifier macro_body
    # macro_body: /.+/
    def macro_definition(self, items):
        identifier = str(items[0])
        macro_body = str(items[1])
        return MacroDefinition(identifier, macro_body)

    def identifier(self, items):
        return items[0]

    def macro_body(self, items):
        return items[0]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

# get current directory
current_dir = os.path.dirname(__file__)
# get package directory
package_dir = os.path.abspath(os.path.join(current_dir, "../../frontend"))
sys.path.insert(0, package_dir)

project = "leap-frontend"
copyright = "2024, Hanyu Wang"
author = "Hanyu Wang"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = []

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
import sphinx_rtd_theme

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"collapse_navigation": False}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


numpydoc_show_class_members = False



from frontend import *
import os
import pytest

# change the working directory to the root directory of the project
os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


# test 00
# compatibility with legup
def test_00_legup():
    verilogDir = "examples/verilogs/external/"
    verilogFile = os.path.join(verilogDir, "legup.v")
    netlist: Netlist = readVerilog(verilogFile)
    module: Module = netlist.getModuleAt(0)

    assert module is not None
    module.exportDOT()
    module.writeDOT("tmp.dot")


# test 01
# compatibility with Google XLS
@pytest.mark.skip(reason="not implemented")
def test_01_xls():
    verilogDir = "examples/verilogs/external/"
    verilogFile = os.path.join(verilogDir, "xls.v")
    netlist: Netlist = readVerilog(verilogFile)
    module: Module = netlist.getModuleAt(0)

    assert module is not None
    module.exportDOT()
    module.writeDOT("tmp.dot")


if __name__ == "__main__":
    test_00_legup()
    test_01_xls()



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 18:07:25
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:55:15
"""

from .dfg import *
from .moduleInst import *
from .ports import *
from .parameters import *


class Module(Frame, ParameterHandler, ExtendedGraph):
    def __init__(
        self,
    ):
        Frame.__init__(self)
        ParameterHandler.__init__(self)
        BNGraph.__init__(self)
        self.setName("top")
        self.submodules = {}
        self.macros = {}

    def addMacro(self, name: str, macro: Any):
        self.macros[name] = macro

    def addInstance(self, inst: ModuleInst):
        self.submodules[inst.getName()] = inst

    def setName(self, module_name: str):
        self.module_name = module_name

    def __repr__(self):
        return f"Module({self.module_name})"

    def loadModuleItems(self, module_items: list):
        # loadModuleItems the DFG
        self.node_is_blocking = {}
        self.submodules = {}
        for item in module_items:
            print(item)

    def getName(self):
        return self.module_name

    def getGraph(self, skipConstants=True) -> pgv.AGraph:
        DotGraph.exportDOT(self, params={"skipConstants": skipConstants})
        return self.graph

    def separateCDFG(self, DFG: pgv.AGraph, CFG: pgv.AGraph):
        fsmGraph = self.graph.add_subgraph(
            DFG,
            name="cluster_data_flow",
            label="DATA_FLOW",
            style="dashed",
            color="blue",
        )
        fsmGraph.graph_attr["rankdir"] = "LR"  # Left-to-right layout

        for node in DFG.nodes():
            self._highlightNode(node, "CYAN")



#
# Copyright (c) 2018-2021 T. {Benz,Kramer}.
#
# This file is part of verilog-parser
# (see https://codeberg.org/tok/py-verilog-parser).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from lark import Lark

# http://www.verilog.com/VerilogBNF.html
# http://www.externsoft.ch/download/verilog.html

from .sv import *
from ..modules import *

GRAMMAR_FILE = "grammar_systemverilog.lark"


def transformVerilogToNetlist(data: str) -> Netlist:
    """
    Parse a string containing data of a verilog file.
    :param data: Raw verilog string.
    :return:
    """
    import os

    curr_dir = os.path.dirname(os.path.abspath(__file__))
    grammar_file = os.path.join(curr_dir, GRAMMAR_FILE)
    with open(grammar_file, "r") as f:
        verilog_netlist_grammar = f.read()

    # select the transformer
    verilog_parser = Lark(
        verilog_netlist_grammar,
        parser="lalr",
        lexer="contextual",
        transformer=SystemVerilogTransformer(),
    )
    netlist = verilog_parser.parse(data)
    return netlist


def parseVerilogToAST(data: str) -> Netlist:
    import os

    curr_dir = os.path.dirname(os.path.abspath(__file__))
    grammar_file = os.path.join(curr_dir, GRAMMAR_FILE)
    with open(grammar_file, "r") as f:
        verilog_netlist_grammar = f.read()

    parser = Lark(verilog_netlist_grammar, parser="earley", lexer="dynamic_complete")
    parseTree = parser.parse(data)

    # return the netlist
    return parseTree


# reference: https://codeberg.org/tok/py-verilog-parser/src/branch/master/verilog_parser/parser.py
def readVerilog(filename: str) -> Netlist:
    """Read the verilog from a file and return the netlist.

    :param filename: the name of the file to read
    :type filename: str
    :raises FileNotFoundError: if the file does not exist
    :return: the netlist
    :rtype: Netlist
    """
    import os

    if not os.path.exists(filename):
        raise FileNotFoundError(f"file {filename} not found")
    with open(filename, "r") as f:
        data = f.read()
    return transformVerilogToNetlist(data)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-28 23:32:17
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-28 23:32:23
"""

from frontend import *
import os

# change the working directory to the root directory of the project
os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


# Test 00
# Check if the test file is correctly initialized
def test_00_Init():
    pass


# Test 01
# Parse the verilog files in the examples/verilogs folder
def test_01_Parser():
    for file in os.listdir("examples/verilogs"):
        if file.endswith(".v"):
            netlist: Netlist = readVerilog("examples/verilogs/" + file)
            assert netlist is not None, "Failed to parse " + file



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 23:20:52
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:22:01
"""

import logging

logger = logging.getLogger(__name__)

from lark import Transformer, v_args, Tree
from ...modules import *

from enum import Enum, auto
from .larkUtils import *


class ConditionTransformer(Transformer):
    def statement(self, items):
        statements = flatten(items)
        for statement in statements:
            assert isinstance(
                statement, Statement
            ), f"statement = {statement}, type = {type(statement)}"
        return statements

    def single_statement(self, items):
        # we make sure that the return value is a list
        return flatten(items)

    def statement_block(self, items):
        return flatten(items)

    # conditional_statement: if_statement else_if_statements? else_statement?
    def conditional_statement(self, items):
        return flatten(items)

    def if_statement(self, items):
        return [x.addCondition(items[0]) for x in items[1]]

    def else_if_statements(self, items):
        return [x.addCondition(items[0]) for x in items[1]]

    def else_statement(self, items):
        return items[0]

    # case_statement: "case" "(" expression ")" case_content "endcase"
    def case_statement(self, items):
        lhs: DFGNode = items[0]
        statements = []
        for cond, statements in items[1:]:
            condition = (
                OPNode("==", OPType.BINARY_EQ, lhs, cond) if cond is not None else None
            )
            statements += [x.addCondition(condition) for x in statements]
        return statements

    def regular_case(self, items) -> list:
        return items

    def default_case(self, items) -> list:
        return [None] + items



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 23:18:52
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-26 00:00:10
"""

from lark import Transformer
from ...modules import *
from .larkUtils import *


class ModuleTransformer(Transformer):

    def netlist(self, items):
        return Netlist(items)

    # this is a list of lots of items, so we set inline=False
    # module: "module" CNAME module_definition ";" declaration* "endmodule"
    def module(self, items):
        name, param_list, port_list = items[:3]
        module: Module = Module()
        module.setName(str(name))
        module.addParameters(param_list)
        module.addPorts(port_list)
        [x(module) for x in flatten(items[3:])]
        return module

    def parameter_list(self, items):
        items = flatten(items)
        return items

    def port_list(self, items):
        items = flatten(items)
        return items

    def port_definition(self, items):
        dir, t, range, name = items
        dir = PortDirection.fromString(dir.data) if dir is not None else None
        t = PortType.fromString(t.data) if t is not None else None
        port = Port(name)
        port.setAll(dir, t, range)
        return port

    def blocking_assignment(self, items):
        return BlockingAssignment(items[0], items[1])

    def non_blocking_assignment(self, items):
        return NonBlockingAssignment(items[0], items[1])

    # module_instantiation: CNAME parameter_instance? CNAME "(" module_instance_arguments ")" ";"
    def module_instantiation(self, items) -> Statement:
        parameter_list = None
        if len(items) == 3:
            module_name, module_inst_name, pin_list = items
        elif len(items) == 4:
            module_name, parameter_list, module_inst_name, pin_list = items
        inst = ModuleInst(module_inst_name)
        inst.setModuleName(module_name)
        if parameter_list is not None:
            inst.setInplaceParameters(parameter_list)
        for key, value in pin_list:
            inst.addPort(key, value)
        return ModuleInstantiation(inst)

    def define_parameter(self, items):
        # usually, this block comes after the module instantiation
        return items



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 17:11:35
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 22:59:35
"""

from .module import Module


class Netlist:
    def __init__(self, modules: list, definitions: dict = {}):
        self.modules = modules[:]

        assert isinstance(definitions, dict), "definitions should be a dict"
        self.definitions = definitions.copy()

    def __repr__(self):
        return "Netlist({})".format(self.modules)

    def getModule(self, name: str) -> Module:
        for module in self.modules:
            if module.module_name == name:
                return module
        return None

    def getModules(self) -> list:
        moduleNames = []
        for module in self.modules:
            moduleNames.append(module.module_name)
        return moduleNames

    def getDefinitions(self) -> dict:
        return self.definitions

    def getModuleAt(self, index: int) -> Module:
        return self.modules[index]



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-06-25 16:57:10
Last Modified by: Hanyu Wang
Last Modified time: 2024-06-25 23:16:24
"""

from .sv import *
from .reader import *



from ..modules import *
from ..writer import *
from ..reader import *


def createModuleFromAssignment(assignment: Assignment, width: int = 6):
    module = Module()
    module.addAssignment(assignment)
    outVar = assignment.target.name
    outPort = OutputPort(outVar, range=Range(width - 1, 0))
    module.addPort(outPort)

    deps = module.getDependencies(outVar)
    for dep in deps:
        module.addPort(InputPort(dep, range=Range(width - 1, 0)))

    return module


def elaborateAssignment(assignment: Assignment):
    module = createModuleFromAssignment(assignment)

    # print module
    tmpFile = "tmp.v"
    writeModule(open(tmpFile, "w"), module)
    import os
    import subprocess

    # get abs path of this file's directory
    path = os.path.dirname(os.path.abspath(__file__))
    subprocess.run(f"{path}/run.sh {tmpFile} {tmpFile} top", shell=True)

    netlist = readVerilog(tmpFile)
    module = netlist.getModule("top")
    module.expandAssignment()

    module.exportDOT()
    module.writeDOT("toy.dot")



from .dotGraph import *
import re


class ExtendedGraph(DotGraph):
    def __init__(self):
        super().__init__()

    def exportDOT(self, varName: str = None, params: dict = {}):
        super().exportDOT(varName, params)

        extractFSM = params.get("extractFSM", True)
        extractCIP = params.get("extractCIP", True)
        extractARG = params.get("extractARG", True)
        extractREG = params.get("extractREG", True)

        if extractFSM:
            # extract the FSM into a subgraph
            fsmGraph = self.graph.add_subgraph(
                name="cluster_fsm", label="FSM", style="dashed", color="blue"
            )
            fsmGraph.graph_attr["rankdir"] = "LR"  # Left-to-right layout

            # add the nodes
            fsmPorts = ["next_state", "finish"]
            self.applyToFrame(
                fsmPorts, lambda node: self._moveToSubgraph(node, self.graph, fsmGraph)
            )
            self.applyToFrame(fsmPorts, lambda node: self._highlightNode(node, "CYAN"))

        if extractCIP:
            # memory address and memory out
            CIPpairs = []
            for node in self.graph.nodes():
                # e.g. main_0_b_address_a and main_0_b_out_a
                if re.match(r".*address.*", node.attr["label"]):
                    try:
                        outNode = self.graph.get_node(
                            node.attr["label"].replace("address", "out")
                        )
                    except KeyError:
                        outNode = None
                    try:
                        inNode = self.graph.get_node(
                            node.attr["label"].replace("address", "in")
                        )
                    except KeyError:
                        inNode = None
                    CIPpairs.append((node, outNode, inNode))

            for i, (addressNode, outNode, inNode) in enumerate(CIPpairs):
                cipGraph = self.graph.add_subgraph(
                    name=f"cluster_cip{i}", label="CIP", style="dashed", color="green"
                )
                cipGraph.graph_attr["rankdir"] = "TB"

                for node in [addressNode, outNode, inNode]:
                    if node is not None:
                        self._moveToSubgraph(node, self.graph, cipGraph)
                        self._highlightNode(node, "GREEN")

        if extractARG:
            argGraph = self.graph.add_subgraph(
                name="cluster_arg", label="args", style="dashed", color="purple"
            )
            node: pgv.Node
            for node in self.graph.nodes():
                if node.attr["label"].startswith("arg"):
                    argGraph.graph_attr["rankdir"] = "LR"
                    self._moveToSubgraph(node, self.graph, argGraph)
                    self._highlightNode(node, "PURPLE")

        if extractREG:
            stages = {}
            for node in self.graph.nodes():
                # match the state number, the name should be ..._reg_stage<num>
                try:
                    stage = int(
                        re.search(r"_reg_stage(\d+)", node.attr["label"]).group(1)
                    )
                    if stage not in stages:
                        stages[stage] = []
                    stages[stage].append(node)
                except AttributeError:
                    continue

            for stage, nodes in stages.items():
                regGraph = self.graph.add_subgraph(
                    name=f"cluster_reg{stage}",
                    label=f"pipeline stage \#{stage}",
                    style="dashed",
                    color="orange",
                )
                regGraph.graph_attr["rankdir"] = "TB"
                for node in nodes:
                    self._moveToSubgraph(node, self.graph, regGraph)
                    self._highlightNode(node, "ORANGE")

        self.graph.layout(prog="dot")



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-23 22:33:09
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-23 23:19:24
"""

from ..modules import *
from .headerWriter import *
from .dfgWriter import *


def moduleHeaderToString(module: Module):
    name = module.getName()
    portString = [f"{port}" for port in module.getIOs()]
    return f"module {name}(\n" + ",\n".join(portString) + "\n);\n\n"


def writeModuleHeader(f, module: Module):
    headerString = moduleHeaderToString(module)
    f.write(headerString)


def parametersToString(parameters: dict):
    retString = ""
    for _, parameter in parameters.items():
        rangeString = rangeToString(parameter.range)
        retString += f"parameter {rangeString}{parameter.name} = {parameter.value};\n"
    return retString + "\n"


def writeParameters(f, module: Module):
    parameters: dict = module.getParameters()
    if len(parameters) > 0:
        f.write(parametersToString(parameters))


def portDefsToString(portDefs: dict):
    retString = ""
    for _, port in portDefs.items():
        retString += portToString(port) + ";\n"
    return retString + "\n"


def writePortDefinitions(f, module: Module):
    portDefs: dict = module.getPorts()
    if len(portDefs) > 0:
        f.write(portDefsToString(portDefs))


def moduleInstToString(moduleInst: ModuleInst):
    inplaceParameterString = ""
    if moduleInst.hasInplaceParameters():
        inplaceParameterList = [str(x) for x in moduleInst.getInplaceParameters()]
        inplaceParameterString = "#(" + ",".join(inplaceParameterList) + ")"

    retString = f"{moduleInst.getModuleName()} {inplaceParameterString} {moduleInst.getInstName()} (\n"
    portStrings = []
    if moduleInst.is_standard_inst:
        ports = moduleInst.getPortList()
        for port in ports:
            # we skip the comma for the last port
            portStrings.append(f"\t.{port} ({moduleInst.getPort(port)})")
    else:
        ports = moduleInst.getPortNames()
        for port in ports:
            portStrings.append(f"\t{port}")
    retString += ",\n".join(portStrings) + "\n"
    retString += ");\n"

    parameters = moduleInst.getParameters()
    parameterStrings = []
    if moduleInst.hasParameters():
        retString += "\n"
        retString += "defparam\n"

        for key, value in parameters.items():
            parameterStrings.append(f"\t{moduleInst.getInstName()}.{key} = {value}")

        retString += ",\n".join(parameterStrings) + ";\n"

        retString += "\n"
    return retString


def writeModuleInst(f, moduleInst: ModuleInst):
    moduleInstString = moduleInstToString(moduleInst)
    f.write(moduleInstString)


def moduleToString(module: Module):
    moduleString = ""
    moduleString += moduleHeaderToString(module)
    moduleString += parametersToString(module.getParameters())
    moduleString += portDefsToString(module.getPorts())
    moduleString += "\n"

    for _, moduleInst in module.submodules.items():
        moduleString += moduleInstToString(moduleInst)

    for var in module.var2assigns:
        for assign in module.getAssignmentsOf(var):
            # TODO: consider the wire/latch/reg
            moduleString += "assign " + assignmentToString(assign) + ";\n"
    moduleString += "endmodule\n"
    return moduleString


def writeModule(f, module: Module):
    writeModuleHeader(f, module)
    writeParameters(f, module)
    writePortDefinitions(f, module)
    f.write("\n")

    for _, moduleInst in module.submodules.items():
        writeModuleInst(f, moduleInst)

    writeAssignments(f, module)
    f.write("endmodule\n")



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-23 23:02:11
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-24 02:14:45
"""

from ..modules import *


def assignmentToString(assignment: Assignment):
    target = assignment.target
    expression = assignment.expression
    condition = assignment.condition

    # TODO: consider wire/latch/reg
    if condition is not None:
        return f"{target} = {expression} ({condition})"
    return f"{target} = {expression}"


def writeAssignment(f, assignment: Assignment):
    target = assignment.target
    expression = assignment.expression
    condition = assignment.condition

    # TODO: consider wire/latch/reg
    f.write("always @(*) begin\n")
    if condition is not None:
        f.write(f"if ({condition}) begin\n")
    f.write(f"\t{target} = {expression};\n")
    if condition is not None:
        f.write("end\n")
    f.write("end\n\n")


def writeAssignments(f, module: Module):
    for var in module.var2assigns:
        for assign in module.getAssignmentsOf(var):
            writeAssignment(f, assign)


def writeModuleDFG(f, module: Module):
    dfg = module.getDFG()
    for node in dfg.getNodes():
        if not node.isVariable:
            continue
        if not node.hasChildren:
            continue

        variableName = node.label
        nodeType = module.getVariableType(node.label)

        # TODO: this could be done more elegantly
        lhs = VariableNode(variableName)
        if node.range is not None:
            lhs.setRange(node.range)

        assert (
            variableName in module.node_is_blocking
        ), f"node {node.label} in module {module.module_name} is not in module.node_is_blocking"
        isBlocking = module.node_is_blocking[variableName]

        children = [dfg.getNode(child) for child in node.children]

        if nodeType == "reg":
            if isBlocking:
                f.write("always @(*) begin\n")
            else:
                f.write("always @(posedge clk) begin\n")

        hasIf: bool = False
        for child in children:
            childType = child.op

            if childType == SOPType.ASSIGN:
                assert (
                    len(child.children) == 1
                ), f"node {child.label} in module {module.module_name} has more than one child: {child.children}"

                if hasIf:
                    f.write("else begin\n")

                childExpression = dfg.extractNodeHelper(child.children[0]).toString()
                if nodeType == "reg":
                    if isBlocking:
                        f.write(f"\t{lhs} = {childExpression};\n")
                    else:
                        f.write(f"\t{lhs} <= {childExpression};\n")
                else:
                    f.write(f"assign {lhs} = {childExpression};\n\n")

                if hasIf:
                    f.write("end\n")

                continue
            if childType == SOPType.CONDITIONAL_ASSIGN:
                assert (
                    len(child.children) == 2
                ), f"node {child.label} in module {module.module_name} has more than two children: {child.children}"

                ifString = "if" if not hasIf else "else if"

                # this means that we have an if statement
                hasIf = True

                condition = dfg.extractNodeHelper(child.children[0]).toString()
                trueExpression = dfg.extractNodeHelper(child.children[1]).toString()

                f.write(f"{ifString} ({condition}) begin\n")
                if nodeType == "reg":
                    if isBlocking:
                        f.write(f"\t{lhs} = {trueExpression};\n")
                    else:
                        f.write(f"\t{lhs} <= {trueExpression};\n")
                else:
                    assert isinstance(
                        lhs, str
                    ), f"variableName's type is {type(variableName)}"
                    f.write(f"\tassign {lhs} = {trueExpression};\n\n")
                f.write("end\n")

        if nodeType == "reg":
            f.write("end\n")

        assert (
            node.op == SOPType.VARIABLE
        ), f"node {node.label} in module {module.module_name} is not a variable"



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2024-07-19 00:28:31
Last Modified by: Hanyu Wang
Last Modified time: 2024-07-23 23:20:41
"""

from ..modules import *
from .headerWriter import *
from .moduleWriter import *


def writeDefinitions(f, netlist: Netlist):
    definitions = netlist.getDefinitions()
    for key, value in definitions.items():
        if key in ["time_unit", "time_precision"]:
            continue
        f.write(f"`define {key} {value}\n")

    if "time_unit" in definitions:
        f.write(
            f"`timescale {definitions['time_unit']} / {definitions['time_precision']}\n\n"
        )


def writeVerilogAST(module: Module, filename: str):
    assert isinstance(module, Module), "module should be an instance of Module"
    module.dfg.toGraph(filename)


def writeVerilog(netlist: Netlist | Module, filename: str):
    """
    Write the netlist to a verilog file

    :param netlist: the netlist or module to write
    :type netlist: Netlist | Module
    :param filename: the name of the file to write
    :type filename: str

    .. note::
        The input can be either a netlist or a module

    Example usage:

    .. code-block:: python

        netlist = readVerilog("path/to/verilog/file.v")
        writeVerilog(netlist, "path/to/output/file.v")

    .. seealso::

        Related functions: :func:`readVerilog`
    """
    if isinstance(netlist, Module):
        writeModule(netlist)
        return

    with open(filename, "w") as f:
        writeHeader(f)
        writeDefinitions(f, netlist)
        for moduleName in netlist.getModules():
            module = netlist.getModule(moduleName)
            writeModule(module)
